from concurrent.futures import TimeoutError, CancelledError
import time
from ..tasks import Mask
from .. import corequest
from skitai import was
from aquests.athreads import trigger
import sys
from ..httpbase.task import DEFAULT_TIMEOUT

class Task (corequest):
    def __init__ (self, future, name, meta, filter, timeout = None):
        self.setup (name, meta, filter, timeout)
        self.future = future

    def setup (self, name, meta, filter, timeout):
        self._name = name
        self._meta = self.meta = meta or {}
        self._filter = filter
        self._started = time.time ()
        self._was = None
        self._fulfilled = None
        self._mask = None
        self._timeout = timeout

    def __str__ (self):
        return self._name

    def __getattr__ (self, name):
        return getattr (self.future, name)

    def get_name (self):
        return self._name

    def _settle (self, future = None):
        if self._fulfilled:
            mask = self._create_mask (0)
            if '__reqid' in self._meta:
                self._fulfilled (mask)
                self._fulfilled = None
            else:
                self._late_respond (self._mask)

    def kill (self):
        try: self.future.result (timeout = 0)
        except: pass
        self.future.set_exception (TimeoutError)

    def cancel (self):
        try: self.future.cancel ()
        except: pass
        self.future.set_exception (CancelledError)

    def set_callback (self, func, reqid = None, timeout = None):
        if reqid is not None:
            self._meta ["__reqid"] = reqid
        self._timeout = timeout
        self._fulfilled = func
        self.future.add_done_callback (self._settle)

    def then (self, func = None, was = None):
        self._fulfilled = func or "self"
        self._was = was or self._get_was ()
        self.future.add_done_callback (self._settle)
        return self

    # common corequest methods ----------------------------------
    def _create_mask (self, timeout):
        self._timeout = timeout
        if self._mask:
            return self._mask
        expt, data = self.future.exception (self._timeout), None
        if not expt:
            data = self.future.result (self._timeout)
            if self._filter:
                data =  self._filter (data)
        self._mask = Mask (data, expt, meta = self.meta)
        return self._mask

    def fetch (self, cache = None, cache_if = (200,), timeout = DEFAULT_TIMEOUT):
        return self._create_mask (timeout).fetch ()

    def one (self, cache = None, cache_if = (200,), timeout = DEFAULT_TIMEOUT):
        return self._create_mask (timeout).one ()

    def commit (self, cache = None, cache_if = (200,), timeout = DEFAULT_TIMEOUT):
        return self._create_mask (timeout).commit ()

    def dispatch (self, cache = None, cache_if = (200,), timeout = DEFAULT_TIMEOUT):
        return self._create_mask (timeout)



import sys
import os
from .commands import smtpda, status, pstats
 
def help ():
    print ("usage: skitai <command> [<options>]")
    print ("command:")
    print ("  smtpda")
    print ("  status")    
    print ("  pstats")
    sys.exit ()
     
def main ():    
    try: 
        cmd = sys.argv [1]
    except IndexError:
        help ()
    sys.argv.pop (1)    
    if cmd == "smtpda":
        smtpda.main ()
    elif cmd == "status":
        status.main ()
    elif cmd == "pstats":
        pstats.main ()    
    else:
        print ("unknown conmand")


if __name__ == "__main__":
    main ()



import multiprocessing
import rs4
from rs4.logger import screen_logger
import time
from .task import Task
import sys

N_CPU = multiprocessing.cpu_count()

class ThreadExecutor:
    NAME = "thread"
    MAINTERN_INTERVAL = 10

    def __init__ (self, workers = None, zombie_timeout = None, logger = None):
        self.logger = logger
        self.workers = workers or N_CPU
        self.zombie_timeout = zombie_timeout
        self.last_maintern = time.time ()
        self.lock = multiprocessing.Lock ()
        self.executor = None
        self.futures = []
        self.no_more_request = False

        self._dones = 0
        self._timeouts = 0

    def __len__ (self):
        with self.lock:
            return len (self.futures)

    def launch_executor (self):
        self.executor = rs4.tpool (self.workers)

    def status (self):
        with self.lock:
            return dict (
                completions = self._dones,
                timeouts = self._timeouts,
                maintainables = len (self.futures),
                zombie_timeout = self.zombie_timeout,
                workers = self.workers,
                last_maintern = self.last_maintern,
                activated = self.executor is not None,
                tasks = [f.get_name () for f in self.futures if not f.done ()]
            )

    def maintern (self, now):
        self.last_maintern = now
        inprogresses = []
        for future in self.futures:
            if future.done ():
                self._dones += 1
                if future.exception ():
                    try:
                        raise future.exception ()
                    except:
                        self.logger.trace ()
                continue

            if self.no_more_request:
                timeout = -1 # kill immediately
            else:
                timeout = future.get_timeout ()
                if timeout is None:
                    timeout = self.zombie_timeout

            # timeout is 0 or None, it is infinite task
            if timeout is not None and future._started + timeout < now:
                future.kill ()
                self._timeouts += 1
                self.logger ("zombie {} task is killed: {}".format (self.NAME, future))
            else:
                inprogresses.append (future)
                continue
        self.futures = inprogresses

    def shutdown (self):
        with self.lock:
            self.no_more_request = True
            if not self.executor:
                return
            self.maintern (time.time ())
            # if False, Py3.7 raise OSError: OSError: handle is closed
            if sys.version_info [:2] >= (3, 10):
                self.executor.shutdown (cancel_futures = True)
            else:
                self.executor.shutdown ()

            self.executor = None
            self.futures = []
            return len (self.futures)

    def __call__ (self, was_id, f, *a, **b):
        with self.lock:
            if self.no_more_request:
                return
            if self.executor is None:
                self.launch_executor ()
            else:
                now = time.time ()
                if now > self.last_maintern + self.MAINTERN_INTERVAL:
                    self.maintern (now)

        meta = {}
        timeout, filter = None, None
        if not a:
            try: meta = b.pop ('meta')
            except KeyError: pass
            try: timeout = b.pop ('timeout')
            except KeyError: pass
            try: filter = b.pop ('filter')
            except KeyError: pass
            try: a = b.pop ('args')
            except KeyError: pass
            b = b.get ('kwargs', b)

        meta ['__was_id'] = was_id
        future = self.executor.submit (f, *a, **b)
        wrap = Task (future, "{}.{}".format (f.__module__, f.__name__), meta = meta, filter = filter)
        timeout and wrap.set_timeout (timeout)
        self.logger ("{} task started: {}".format (self.NAME, wrap))
        with self.lock:
            self.futures.append (wrap)
        return wrap

class ProcessExecutor (ThreadExecutor):
    NAME = "process"
    def launch_executor (self):
        self.executor = rs4.ppool (self.workers)


# ------------------------------------------------------------------------

class Executors:
    def __init__ (self, workers = N_CPU, zombie_timeout = None, logger = None):
        self.logger = logger or screen_logger ()
        self.executors = [
            ThreadExecutor (workers, zombie_timeout, self.logger),
            ProcessExecutor (workers, zombie_timeout, self.logger)
        ]

    def status (self):
        return dict (
            thread = self.executors [0].status (),
            process = self.executors [1].status ()
        )

    def cleanup (self):
        if sys.version_info [:2] >= (3, 10):
            return [e.shutdown (cancel_futures = True) for e in self.executors]
        else:
            return [e.shutdown () for e in self.executors]

    def create_thread (self, was_id, f, *a, **b):
        return self.executors [0] (was_id, f, *a, **b)

    def create_process (self, was_id, f, *a, **b):
        return self.executors [1] (was_id, f, *a, **b)




import time, os
from rs4.psutil import processutil
from rs4.termcolor import tc

def due (sec):
    if sec < 60:
        return "{} seconds".format (int (sec))
    elif sec < 3600:
        return "{} minutes".format (sec // 60)
    elif sec < 86400:
        return "{:.1f} hours".format (sec / 3600)
    return "{:.1f} days".format (sec / 3600 / 24)

def main ():
    print ("existing apps:")
    for each in sorted (os.listdir ("/var/tmp/skitai")):
        if each.startswith ('_'):
            continue
        if os.path.isdir (os.path.join ("/var/tmp/skitai", each)):
            plock = os.path.join ("/var/tmp/skitai", each, ".pid")
            running = False
            if os.path.isfile (plock):
                with open (plock) as f:
                    pid = int (f.read ())
                running = processutil.is_running (pid, "skitai/")

        if running:
            inst = tc.debug (each)
            status = "running for {}, pid {}".format (tc.debug (due (time.time () - os.path.getmtime (plock))), tc.debug (pid))
        else:
            inst = tc.primary (each)
            status = "stopped"
        print ("- {} {}".format (inst, status))
    print ("\nif you want to remove app from list,")
    print ("  rm -rf /var/tmp/skitai/<app>")
